/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

#ifndef _PORTABLE_QUEUE_H
#define _PORTABLE_QUEUE_H

#include <new.h>
#include "support/PortableTypeDefs.h"
#include "support/PortableMacros.h"

//: This class is very similar to an STL vector, Java Vector, or BList,
//: except that it is implemented in such a way that adding or removing
//: items from the head or tail of the list is an O(1) operation.
template <class ItemType> class PortableQueue
{
public:
   // (initialSize) specifies how many slots to pre-allocate.  Defaults to five.
   PortableQueue(uint32 initialSize = 5);

   PortableQueue(const PortableQueue& copyMe);
   virtual ~PortableQueue();

   PortableQueue &operator=(const PortableQueue &from);

   // Puts (item) at the end of the queue.  Queue size grows by one.
   // Returns true on success, false on failure (i.e. out of memory)
   bool AddTail(const ItemType &item);

   // Puts (item) at the head of the queue.  Queue size grows by one.
   // Returns true on success, false on failure (i.e. out of memory)
   bool AddHead(const ItemType &item);

   // Removes the item at the head of the queue and places it into (returnItem).
   // Returns true on success, false on failure (i.e. queue was empty)
   bool RemoveHead(ItemType & returnItem);

   // Removes the item at the head of the queue.
   // Returns true on success, false on failure (i.e. queue was empty)
   bool RemoveHead();

   // Removes the item at the tail of the queue and places it into (returnItem).
   // Returns true on success, false on failure (i.e. queue was empty)
   bool RemoveTail(ItemType & returnItem);

   // Removes the item at the tail of the queue.
   // Returns true on success, false on failure (i.e. queue was empty)
   bool RemoveTail();

   // Removes the item at the (index)'th position in the queue.
   // (index) can range from zero (head of the queue) to 
   // (CountItems()-1) (tail of the queue).
   // Returns true on success, false on failure (i.e. bad index)
   bool RemoveItemAt(uint32 index);

   // Removes the item at the (index)'th position in the queue,
   // and places it into (returnItem).
   // (index) can range from zero (head of the queue) to 
   // (CountItems()-1) (tail of the queue).
   // Returns true on success, false on failure (i.e. bad index)
   bool RemoveItemAt(uint32 index, ItemType & returnItem);

   // Copies the (index)'th item into (returnItem).
   // Returns true on success, false on failure (e.g. bad index)
   bool GetItemAt(uint32 index, ItemType & returnItem) const;

   // Replaces the (index)'th item in the queue with (newItem).
   // Returns true on success, false on failure (e.g. bad index)
   bool ReplaceItemAt(uint32 index, const ItemType & newItem);
 
   // Removes all items from the queue.
   void MakeEmpty();

   // Returns the number of items in the queue.  (This number
   // does not include additional pre-allocated space)
   uint32 CountItems() const {return _itemCount;}

   // Returns true iff the queue is empty.
   bool IsEmpty() const {return (_itemCount == 0);}

   // Convenient array-style get operator (be sure to only use valid indices!)
   ItemType operator [](uint32 Index) const; 

   // Convenient array-style set operator (be sure to only use valid indices!)
   ItemType & operator [](uint32 Index);

   // Returns a pointer to an item in the array (i.e. no 
   // copying of the item is done).  Included for efficiency;
   // be careful with this, modifying the queue can invalidate 
   // the returned pointer!
   ItemType * GetItemPointer(uint32 index) const;

   // Makes sure there is enough space allocated for at least
   // (numItems) items.  Returns true on success, false on
   // failure (out of memory).  You only need to call this if
   // you wish to minimize the number of data re-allocations done.
   bool EnsureSize(uint32 numItems);

   // Returns the index of the given (item), or -1 if (item) is
   // not found in the list.  O(n) time.
   int32 IndexOf(const ItemType & item) const;

private:
   int32 NextIndex(int32 idx) const {return (idx >= _queueSize-1) ? 0 : idx+1;}
   int32 PrevIndex(int32 idx) const {return (idx <= 0) ? _queueSize-1 : idx-1;}

   // Translates a user-index into an index into the _queue array.
   uint32 InternalizeIndex(uint32 idx) const {return (_headIndex + idx) % _queueSize;}

   ItemType * _queue; 
   uint32 _queueSize;  // number of slots in the _queue array
   uint32 _itemCount;  // number of valid items in the array
   int32 _headIndex;   // index of the first filled slot, or -1 
   int32 _tailIndex;   // index of the last filled slot, or -1
   const uint32 _initialSize;  // as specified in ctor

   // used to "reset" slots, which will free memory when using PortableRefs
   const ItemType _blankItem;  
};

template <class ItemType>
PortableQueue<ItemType>::PortableQueue(uint32 initialSize)
   : _queue(NULL), _queueSize(0), _initialSize(initialSize), _blankItem(), _itemCount(0)
{
   MakeEmpty();
}

template <class ItemType>
PortableQueue<ItemType>::PortableQueue(const PortableQueue& rhs)
   : _queue(NULL), _queueSize(0), _initialSize(rhs._initialSize), _blankItem(), _itemCount(0)
{
   MakeEmpty();
   *this = rhs;
}


template <class ItemType>
PortableQueue<ItemType> &
PortableQueue<ItemType>::operator =(const PortableQueue& rhs)
{
   if (this != &rhs)
   {
      MakeEmpty();
      uint32 numItems = rhs.CountItems();
      if (EnsureSize(numItems))
      {
         ItemType temp;
         for (uint32 i=0; i<numItems; i++) 
         {
            if ((rhs.GetItemAt(i,temp) == false)||(AddTail(temp) == false)) WARN_OUT_OF_MEMORY;
         }
      }
   }
   return *this;
}

template <class ItemType>
ItemType &
PortableQueue<ItemType>::operator[](uint32 i)
{
   ASSERT(((i>=0)&&(i<_itemCount)), "Invalid index to PortableQueue::[]");
   return _queue[InternalizeIndex(i)];
}

template <class ItemType>
ItemType
PortableQueue<ItemType>::operator[](uint32 i) const
{
   ASSERT(((i>=0)&&(i<_itemCount)), "Invalid index to PortableQueue::[]");
   return _queue[InternalizeIndex(i)];
}             

template <class ItemType>
ItemType * 
PortableQueue<ItemType>::GetItemPointer(uint32 i) const
{
   return &_queue[InternalizeIndex(i)];
}

template <class ItemType>
PortableQueue<ItemType>::~PortableQueue()
{
   delete [] _queue;
}


template <class ItemType>
bool 
PortableQueue<ItemType>::
AddTail(const ItemType &item)
{
   if (EnsureSize(_itemCount+1) == false) return false;
   if (_itemCount == 0) _headIndex = _tailIndex = 0;
                   else _tailIndex = NextIndex(_tailIndex);
   _queue[_tailIndex] = item;
   _itemCount++;
   return true;
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
AddHead(const ItemType &item)
{
   if (EnsureSize(_itemCount+1) == false) return false;
   if (_itemCount == 0) _headIndex = _tailIndex = 0;
                   else _headIndex = PrevIndex(_headIndex);
   _queue[_headIndex] = item;
   _itemCount++;
   return true;
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
RemoveHead(ItemType & returnItem)
{
   if (_itemCount == 0) return false;
   returnItem = _queue[_headIndex];
   return RemoveHead();
}

template <class ItemType>
bool
PortableQueue<ItemType>::
RemoveHead()
{
   if (_itemCount == 0) return false;
   _queue[_headIndex] = _blankItem;
   _headIndex = NextIndex(_headIndex);
   _itemCount--;
   return true;
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
RemoveTail(ItemType & returnItem)
{
   if (_itemCount == 0) return false;
   returnItem = _queue[_tailIndex];
   return RemoveTail();
}


template <class ItemType>
bool 
PortableQueue<ItemType>::
RemoveTail()
{
   if (_itemCount == 0) return false;
   _queue[_tailIndex] = _blankItem;
   _tailIndex = PrevIndex(_tailIndex);
   _itemCount--;
   return true;
}

template <class ItemType>
bool
PortableQueue<ItemType>::
GetItemAt(uint32 index, ItemType & returnItem) const
{
   if (index >= _itemCount) return false;
   returnItem = _queue[InternalizeIndex(index)];
   return true;
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
RemoveItemAt(uint32 index, ItemType & returnItem)
{
   if (index >= _itemCount) return false;
   returnItem = _queue[InternalizeIndex(index)];
   return RemoveItemAt(index);
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
RemoveItemAt(uint32 index)
{
   if (index >= _itemCount) return false;
   if (index == 0) return RemoveHead();  // this is cheaper!

   index = InternalizeIndex(index);
   while(index != _tailIndex)
   {
      uint32 next = NextIndex(index);
      _queue[index] = _queue[next];
      index = next;
   }
   _queue[_tailIndex] = _blankItem;
   _tailIndex = PrevIndex(_tailIndex); 
   _itemCount--;
   return true; 
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
ReplaceItemAt(uint32 index, const ItemType & newItem)
{
   if (index >= _itemCount) return false;
   _queue[InternalizeIndex(index)] = newItem;
   return true;
}

template <class ItemType>
void 
PortableQueue<ItemType>::
MakeEmpty()
{
   while(_itemCount > 0) RemoveTail();
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
EnsureSize(uint32 size)
{
   if ((_queue == NULL)||(_queueSize < size))
   {
      uint32 newQLen = size*2;
      if (newQLen < _initialSize) newQLen = _initialSize;

      ItemType * newQueue = new (nothrow) ItemType[newQLen];
      if (newQueue == NULL) {WARN_OUT_OF_MEMORY; return false;}
      
      if (_itemCount > 0)
      {
         for (int i=0; i<_itemCount; i++) (void) GetItemAt(i, newQueue[i]);
         _headIndex = 0;
         _tailIndex = _itemCount-1;
      }

      delete [] _queue;
      _queue = newQueue;
      _queueSize = newQLen;
   }
   return true;
}

template <class ItemType>
int32 
PortableQueue<ItemType>::
IndexOf(const ItemType & item) const
{
   if (_queue) for (int i=CountItems()-1; i>=0; i--) if (*GetItemPointer(i) == item) return i;
   return -1;
}

#endif

